from collections.abc import Iterable

from ...errors import MissingDiscordLibrary

try:
    from .dc_imports import discord
except ImportError:
    raise MissingDiscordLibrary()

commands = __import__(f"{discord.lib}.ext.commands", fromlist=[""])
tasks = __import__(f"{discord.lib}.ext.tasks", fromlist=[""])


try:
    # py-cord
    from discord import CogMeta
    from discord.ext import bridge
    from discord.utils import AutocompleteFunc, V, Values

except ImportError:
    CogMeta = commands.CogMeta
    bridge = commands

    AutocompleteFunc = None
    V = Iterable[str]
    Values = Iterable[str]



"""Embed templates that can be used to send messages.
These functions will generate embeds and send them to the desired target.

Example
-------
.. code-block:: python

    from ezcord import Bot, emb

    bot = Bot()

    @bot.slash_command()
    async def hey(ctx):
        await emb.success(ctx, "Success!")
"""

from __future__ import annotations

import copy

from .internal import copy_kwargs, load_embed, replace_dict, save_embeds
from .internal.dc import discord
from .logs import log


def set_embed_templates(
    *,
    error_embed: discord.Embed | str | None = None,
    success_embed: discord.Embed | str | None = None,
    warn_embed: discord.Embed | str | None = None,
    info_embed: discord.Embed | str | None = None,
    **kwargs: discord.Embed,
):
    """Override the default embeds with custom ones.

    This must be called before the first embed template is used. The description
    of the embeds will be replaced with the given text.

    If you pass a string, error messages will be sent as a text instead of an embed.
    If the string is empty, the text will be taken from template methods.

    .. note::
        You can use the following variables for embed templates. They will automatically
        be replaced when the template is sent to an interaction.

        - ``{user}`` - The user who initiated the interaction
        - ``{username}`` - The name of the user
        - ``{user_mention}`` - The user mention
        - ``{user_id}`` - The ID of the user
        - ``{user_avatar}`` - The URL of the user's avatar

        Server variables will be replaced with information about the bot if the interaction
        was initiated in DMs.

        - ``{servername}`` - The guild where the interaction was initiated
        - ``{server_icon}`` - The URL of the guild's icon

    Parameters
    ----------
    error_embed:
        The embed to use for error messages.
    success_embed:
        The embed to use for success messages.
    warn_embed:
        The embed to use for warning messages.
    info_embed:
        The embed to use for info messages.
    **kwargs:
        Additional embed templates. Can be used with :func:`send`.

    Example
    -------
    .. code-block:: python

        from ezcord import emb

        embed = discord.Embed(
            title="Error",
            color=discord.Color.orange()
        )
        emb.set_embed_templates(error_embed=embed)
    """
    save_embeds(
        error_embed=error_embed,
        success_embed=success_embed,
        warn_embed=warn_embed,
        info_embed=info_embed,
        **kwargs,
    )


async def _send_embed(
    target: discord.ApplicationContext | discord.Interaction | discord.abc.Messageable,
    embed: discord.Embed | str,
    ephemeral: bool = True,
    edit: bool = False,
    **kwargs,
):
    """Respond to an interaction or send an embed to a target.

    If the interaction has already been responded to,
    the message will be sent as a followup.

    Parameters
    ----------
    target:
        The application context or the interaction to send the message to.
    embed:
        The embed to send.
    ephemeral:
        Whether the message should be ephemeral. Defaults to ``True``.
    """
    content = None
    if isinstance(embed, str):
        content = embed
        embed = None

    if isinstance(target, discord.ApplicationContext) or isinstance(target, discord.Interaction):
        if edit:
            try:
                await target.edit(content=content, embed=embed, **kwargs)
                return
            except AttributeError:
                log.error("'edit=True' can only be used with Pycord master branch.")

        if not target.response.is_done():
            await target.response.send_message(
                content=content, embed=embed, ephemeral=ephemeral, **kwargs
            )
        else:
            await target.followup.send(content=content, embed=embed, ephemeral=ephemeral, **kwargs)
    else:
        await target.send(content=content, embed=embed, **kwargs)


def _insert_info(
    target: discord.ApplicationContext | discord.Interaction | discord.abc.Messageable,
    embed: discord.Embed | str,
):
    if not isinstance(target, discord.ApplicationContext) and not isinstance(
        target, discord.Interaction
    ):
        return embed

    interaction = target
    if isinstance(target, discord.ApplicationContext):
        interaction = target.interaction

    if isinstance(embed, discord.Embed):
        embed = embed.to_dict()
    embed_dic = replace_dict(embed, interaction)

    if isinstance(embed, dict):
        return discord.Embed.from_dict(embed_dic)

    return embed_dic


async def _process_message(
    target: discord.ApplicationContext | discord.Interaction | discord.abc.Messageable,
    embed: discord.Embed | str,
    txt: str | None,
    title: str | None,
    edit: bool,
    ephemeral: bool,
    **kwargs,
):
    if isinstance(embed, discord.Embed):
        embed = copy.deepcopy(embed)
        if txt is not None:
            embed.description = txt
        if title is not None:
            embed.title = title
    elif isinstance(embed, str) and embed == "":
        embed = txt

    embed = _insert_info(target, embed)

    await _send_embed(target, embed, ephemeral, edit, **kwargs)


@copy_kwargs(discord.InteractionResponse.send_message)
async def error(
    target: discord.ApplicationContext | discord.Interaction | discord.abc.Messageable,
    txt: str | None = None,
    *,
    title: str | None = None,
    edit: bool = False,
    ephemeral: bool = True,
    **kwargs,
):
    """Send an error message. By default, this is a red embed.

    Parameters
    ----------
    target:
        The target to send the message to.
    txt:
        The text for the embed description. If this is ``None``,
        you need to provide a non-empty ``Embed`` when using :func:`set_embed_templates`.
    title:
        The title of the embed. Defaults to ``None``.
    edit:
        Whether to edit the last message instead of sending a new one. Defaults to ``False``.
    ephemeral:
        Whether the message should be ephemeral. Defaults to ``True``.
    """
    embed = load_embed("error_embed")
    await _process_message(target, embed, txt, title, edit, ephemeral, **kwargs)


@copy_kwargs(discord.abc.Messageable.send)
async def success(
    target: discord.ApplicationContext | discord.Interaction | discord.abc.Messageable,
    txt: str | None = None,
    *,
    title: str | None = None,
    edit: bool = False,
    ephemeral: bool = True,
    **kwargs,
):
    """Send a success message. By default, this is a green embed.

    Parameters
    ----------
    target:
        The target to send the message to.
    txt:
        The text for the embed description. If this is ``None``,
        you need to provide a non-empty ``Embed`` when using :func:`set_embed_templates`.
    title:
        The title of the embed. Defaults to ``None``.
    edit:
        Whether to edit the last message instead of sending a new one. Defaults to ``False``.
    ephemeral:
        Whether the message should be ephemeral. Defaults to ``True``.
    """
    embed = load_embed("success_embed")
    await _process_message(target, embed, txt, title, edit, ephemeral, **kwargs)


@copy_kwargs(discord.abc.Messageable.send)
async def warn(
    target: discord.ApplicationContext | discord.Interaction | discord.abc.Messageable,
    txt: str | None = None,
    *,
    title: str | None = None,
    edit: bool = False,
    ephemeral: bool = True,
    **kwargs,
):
    """Send a warning message. By default, this is a golden embed.

    Parameters
    ----------
    target:
        The target to send the message to.
    txt:
        The text for the embed description. If this is ``None``,
        you need to provide a non-empty ``Embed`` when using :func:`set_embed_templates`.
    title:
        The title of the embed. Defaults to ``None``.
    edit:
        Whether to edit the last message instead of sending a new one. Defaults to ``False``.
    ephemeral:
        Whether the message should be ephemeral. Defaults to ``True``.
    """
    embed = load_embed("warn_embed")
    await _process_message(target, embed, txt, title, edit, ephemeral, **kwargs)


@copy_kwargs(discord.abc.Messageable.send)
async def info(
    target: discord.ApplicationContext | discord.Interaction | discord.abc.Messageable,
    txt: str | None = None,
    *,
    title: str | None = None,
    edit: bool = False,
    ephemeral: bool = True,
    **kwargs,
):
    """Send an info message. By default, this is a blue embed.

    Parameters
    ----------
    target:
        The target to send the message to.
    txt:
        The text for the embed description. If this is ``None``,
        you need to provide a non-empty ``Embed`` when using :func:`set_embed_templates`.
    title:
        The title of the embed. Defaults to ``None``.
    edit:
        Whether to edit the last message instead of sending a new one. Defaults to ``False``.
    ephemeral:
        Whether the message should be ephemeral. Defaults to ``True``.
    """
    embed = load_embed("info_embed")
    await _process_message(target, embed, txt, title, edit, ephemeral, **kwargs)


@copy_kwargs(discord.abc.Messageable.send)
async def send(
    template: str,
    target: discord.ApplicationContext | discord.Interaction | discord.abc.Messageable,
    txt: str | None = None,
    *,
    title: str | None = None,
    edit: bool = False,
    ephemeral: bool = True,
    **kwargs,
):
    """Send a custom embed template. This needs to be set up with :func:`set_embed_templates`.

    Parameters
    ----------
    template:
        The name of the template that was used in :func:`set_embed_templates`.
    target:
        The target to send the message to.
    txt:
        The text for the embed description. If this is ``None``,
        you need to provide a non-empty ``Embed`` when using :func:`set_embed_templates`.
    title:
        The title of the embed. Defaults to ``None``.
    edit:
        Whether to edit the last message instead of sending a new one. Defaults to ``False``.
    ephemeral:
        Whether the message should be ephemeral. Defaults to ``True``.
    """
    embed = load_embed(template)
    await _process_message(target, embed, txt, title, edit, ephemeral, **kwargs)



"""
Commands to manage the bot blacklist.

This file should only be called through `bot.add_blacklist()`,
after the "blacklist" value has been set in the config.
"""

from typing import Union

import aiosqlite

from .. import emb
from ..blacklist import _BanDB
from ..bot import Bot, Cog
from ..components import event
from ..errors import Blacklisted
from ..internal import EzConfig, t
from ..internal.dc import commands, discord
from ..utils import create_text_file

_db = _BanDB()


async def _check_blacklist(
    # Pycord does not support __future__ in cogs for whatever reason
    interaction: Union[discord.Interaction, discord.ApplicationContext]
) -> bool:
    bans = await _db.get_bans()
    if interaction.user.id in bans and EzConfig.blacklist:
        if EzConfig.blacklist.raise_error:
            raise Blacklisted()
        else:
            await interaction.respond(t("no_perms"), ephemeral=True)
        return False
    return True


@event
async def view_check(interaction: discord.Interaction):
    return await _check_blacklist(interaction)


class Blacklist(Cog, hidden=True):
    def __init__(self, bot: Bot):
        super().__init__(bot)

    @staticmethod
    async def bot_check(ctx: discord.ApplicationContext):
        return await _check_blacklist(ctx)

    admin = discord.SlashCommandGroup(
        t("admin_group"),
        guild_ids=EzConfig.admin_guilds,
        default_member_permissions=discord.Permissions(administrator=True),
        checks=[commands.is_owner().predicate],
    )

    @Cog.listener()
    async def on_guild_join(self, guild: discord.Guild):
        bans = await _db.get_bans()
        if guild.owner.id in bans:
            try:
                await guild.owner.send(t("guild_error", guild.name))
            except discord.Forbidden:
                pass
            await guild.leave()

    blacklist = admin.create_subgroup("blacklist")

    @blacklist.command(name="manage", description="Manage the blacklist")
    @discord.option("choice", description="Choose an action", choices=["Add ban", "Remove ban"])
    @discord.option("user", description="The user to ban/unban")
    @discord.option("reason", description="The reason for the ban", default=None)
    async def manage_blacklist(
        self,
        ctx: discord.ApplicationContext,
        choice: str,
        user: discord.Member,
        reason: str,
    ):
        if choice == "Add ban":
            if user.id == ctx.author.id:
                return await emb.error(ctx, "You can't ban yourself.")
            if user.bot:
                return await emb.error(ctx, "You can't ban a bot.")

            try:
                await _db.add_ban(user.id, reason)
            except aiosqlite.IntegrityError:
                return await emb.error(ctx, "This user is already banned.")
            await ctx.respond(
                f"The user was banned successfully.\n- **Name:** {user}\n- **ID:** {user.id}",
                ephemeral=True,
            )
        else:
            rowcount = await _db.remove_ban(user.id)
            if rowcount == 0:
                return await emb.error(ctx, "This user is not banned.")
            await ctx.respond(f"The user **{user}** was unbanned successfully.", ephemeral=True)

    @blacklist.command(name="show", description="Show the bot blacklist")
    async def show_blacklist(self, ctx):
        await ctx.defer(ephemeral=True)
        bans = await _db.get_full_bans()
        desc = ""

        for user_id, reason, _ in bans:
            if not reason:
                reason = "No reason provided"

            user = await self.bot.get_or_fetch_user(user_id)
            name = f"{user.name} ({user.id})" if user else user_id
            desc += f"{name} - {reason}\n"

        if not desc:
            desc = "No bans found."

        file = create_text_file(desc, "bans.txt")
        await ctx.respond(file=file, ephemeral=True)

    @admin.command(description="Show all bot servers")
    async def show_servers(self, ctx):
        await ctx.defer(ephemeral=True)
        longest_name = max([guild.name for guild in self.bot.guilds], key=len)
        sep = f"<{len(longest_name)}"

        desc = ""
        for guild in self.bot.guilds:
            desc += f"{guild.name:{sep}} - {guild.member_count:<6,}"
            desc += f" - {guild.id}"
            desc += f" - {guild.owner} ({guild.owner.id})"
            desc += "\n"

        file = create_text_file(desc, "guilds.txt")
        await ctx.respond(file=file, ephemeral=True)

    leave = admin.create_subgroup("leave")

    @leave.command(name="server", description="Make the bot leave a server")
    @discord.option("guild_id", description="Leave the server with the given ID", default=None)
    async def leave_guild(
        self,
        ctx: discord.ApplicationContext,
        guild_id: str,
    ):
        await ctx.defer(ephemeral=True)
        try:
            guild = await self.bot.fetch_guild(guild_id)
        except Exception as e:
            return await ctx.respond(f"I could not load this server: ```{e}```", ephemeral=True)

        await guild.leave()
        await ctx.respond(f"I left **{guild.name}** ({guild.id})", ephemeral=True)

    @leave.command(name="owner", description="Make the bot leave all guilds with a given owner")
    @discord.option("owner_id", description="Leave all servers with the specified owner")
    async def leave_owner(
        self,
        ctx: discord.ApplicationContext,
        owner: discord.User,
    ):
        await ctx.defer(ephemeral=True)
        guilds = []
        member_count = 0
        for guild in self.bot.guilds:
            if guild.owner.id == owner.id:
                guilds.append(guild)
                member_count += guild.member_count

        return await ctx.respond(
            f"I found **{len(guilds)}** servers with **{owner}** as the owner "
            f"(with a total of **{member_count}** members).",
            ephemeral=True,
            view=LeaveGuilds(guilds),
        )


def setup(bot: Bot):
    bot.add_cog(Blacklist(bot))


class LeaveGuilds(discord.ui.View):
    def __init__(self, guilds: list[discord.Guild]):
        self.guilds = guilds
        super().__init__(timeout=None)

    @discord.ui.button(label="Leave all", style=discord.ButtonStyle.red)
    async def leave(self, _: discord.ui.Button, interaction: discord.Interaction):
        for child in self.children:
            child.disabled = True
        embed = discord.Embed(
            description="I'll now leave all servers. This may take a while."
            "\n\nI'll ping you when I'm done.",
        )
        await interaction.edit(embed=embed, view=self)

        leave_count = 0
        for guild in self.guilds:
            try:
                await guild.leave()
                leave_count += 1
            except Exception as e:
                print(f"Could not leave guild {guild.id}: ```\n{e}```")
                continue

        await interaction.respond(
            f"{interaction.user.mention} I successfully left **{leave_count}** servers.",
            ephemeral=True,
        )

    @discord.ui.button(label="Cancel", style=discord.ButtonStyle.grey)
    async def cancel(self, _: discord.ui.Button, interaction: discord.Interaction):
        for child in self.children:
            child.disabled = True
        await interaction.edit(view=self)



"""Some logging utilities that are used for bot logs."""

from __future__ import annotations

import asyncio
import logging
import os
import re
import sys
from typing import Literal

import aiohttp
from colorama import Fore

from .enums import LogFormat, TimeFormat
from .internal.colors import (
    DEFAULT_COLOR,
    DEFAULT_LOG_COLORS,
    get_escape_code,
    replace_dc_format,
)
from .internal.dc import discord

DEFAULT_LOG = "ezcord"
log = logging.getLogger(DEFAULT_LOG)


def custom_log(
    key: str, message: str, *, color: str | None = DEFAULT_COLOR, level: int = logging.INFO
):
    """Log a message with a custom log level.

    Parameters
    ----------
    key:
        The name of the custom log level.
    message:
        The message to log.
    color:
        The color to use for the log level. Defaults to ``Fore.MAGENTA``.
    level:
        The log level. Defaults to ``logging.INFO``.
    """
    color = get_escape_code(color)
    logging.getLogger(DEFAULT_LOG).log(level, message, extra={"key": key, "color": color})


def _format_log_colors(log_format: str, file: bool, final_colors: dict[int, str]) -> dict[int, str]:
    """Checks if the is sent to a file and formats the colors accordingly."""
    format_colors = re.findall(r"{.*?}", log_format)

    remove_attrs = ["{color}"]
    for attr in remove_attrs:
        format_colors = [x for x in format_colors if attr not in x]

    color_formats = {}
    if "{end}" in log_format:
        for level in final_colors:
            if file:
                for substring in format_colors:
                    log_format = log_format.replace(substring, "")

                color_formats[level] = log_format.format(color="", end="")
            else:
                for substring in format_colors:
                    if substring == "{end}":
                        log_format = log_format.replace(substring, Fore.RESET, 1)
                        continue

                    log_format = log_format.replace(
                        substring, get_escape_code(substring.strip("{}"))
                    )

                color_formats[level] = log_format.format(color=final_colors[level], end=Fore.RESET)
    else:
        for level in final_colors:
            if file:
                color_formats[level] = log_format
            else:
                color_formats[level] = final_colors[level] + log_format + Fore.RESET

    return color_formats


def _format_colors(colors: dict[int, str] | str | None = None) -> dict[int, str]:
    """Overwrite the default colors for the given log levels in the given format."""

    final_colors = DEFAULT_LOG_COLORS.copy()
    if colors is None:
        colors = final_colors

    if isinstance(colors, str):
        colors = get_escape_code(colors)
        for level in final_colors:
            final_colors[level] = colors
    else:
        for level in colors:
            final_colors[level] = get_escape_code(colors[level])

    return final_colors


class _ColorFormatter(logging.Formatter):
    """A logging formatter that adds colors to the output. This is used by :func:`set_log`."""

    def __init__(
        self,
        file: bool,
        log_format: str,
        time_format: str,
        dc_codeblocks: bool,
        spacing: int,
        space_after_level: bool,
        colors: dict[int, str] | str | None = None,
        *args,
        **kwargs,
    ):
        super().__init__(*args, **kwargs)

        self.file = file
        self.colors = colors
        self.dc_codeblocks = dc_codeblocks
        self.spacing = spacing
        self.space_after_level = space_after_level
        self.LOG_FORMAT = log_format
        self.TIME_FORMAT = time_format

    def format(self, record: logging.LogRecord):
        """Adds colors to log messages and formats them accordingly.
        Can be used with :func:`set_log`.

        Parameters
        ----------
        record:
            The log record to format.
        """
        if "color" in record.__dict__:
            colors = record.__dict__["color"]
        else:
            colors = self.colors

        color_formats = _format_colors(colors)

        if record.levelno not in color_formats:
            if "color" in record.__dict__:
                # if a custom log level is used by .custom_log()
                color_formats[record.levelno] = colors
            else:
                # if no color is set for a custom log level used by .log()
                color_formats[record.levelno] = DEFAULT_COLOR

        color_log_formats = _format_log_colors(self.LOG_FORMAT, self.file, color_formats)

        log_format = color_log_formats.get(record.levelno)
        spacing_name = record.__dict__["key"] if "key" in record.__dict__ else record.levelname
        spaces = " " * (self.spacing - len(spacing_name))

        if self.space_after_level:
            # add space directly after log level
            spacing_name = spacing_name + spaces
            record.levelname = spacing_name
        else:
            # add space in front of log message
            record.msg = spaces + record.msg

        if "key" in record.__dict__ and log_format:
            log_format = log_format.replace("%(levelname)s", spacing_name)

        current_level_color = color_formats.get(record.levelno)
        new_record = logging.makeLogRecord(record.__dict__)

        # color new lines
        if isinstance(log_format, str) and log_format.endswith("//"):
            log_format = log_format.replace("//", "")
            split = new_record.msg.split("\n", 1)
            if len(split) > 1:
                new_record.msg = split[0] + "\n***" + split[1] + "***"

        remove_stars = self.file or "{end}" not in self.LOG_FORMAT
        new_record.msg = replace_dc_format(new_record.msg, current_level_color, remove_stars)

        formatter = logging.Formatter(log_format, self.TIME_FORMAT)
        return formatter.format(new_record)


class _DiscordHandler(logging.Handler):
    """A logging handler that sends logs to a Discord webhook."""

    def __init__(
        self, webhook_url: str | None, dc_codeblocks: bool, dc_format: str, *args, **kwargs
    ):
        super().__init__(*args, **kwargs)
        self.webhook_url = webhook_url
        self.dc_codeblocks = dc_codeblocks
        self.dc_format = dc_format

    def emit(self, record: logging.LogRecord):
        """Removes exception information and sets the log format."""
        record._exc_info_hidden, record.exc_info = record.exc_info, None
        record.exc_text = None

        if self.dc_codeblocks:
            msg = f"```ansi\n{self.format(record)}```"
        else:
            if "key" in record.__dict__:
                log_format = self.dc_format.replace("%(levelname)s", record.__dict__["key"])

                formatter = logging.Formatter(log_format)
                msg = formatter.format(record)
            else:
                msg = self.format(record)

        if self.webhook_url:
            loop = asyncio.get_event_loop()
            if loop.is_running():
                loop.create_task(_send_discord_log(self.webhook_url, record, msg))
            else:
                loop.run_until_complete(_send_discord_log(self.webhook_url, record, msg))


async def _send_discord_log(webhook_url: str, record: logging.LogRecord, msg):
    async with aiohttp.ClientSession() as session:
        webhook = discord.Webhook.from_url(webhook_url, session=session)
        name = "EzCord" if record.name == DEFAULT_LOG else record.name
        try:
            await webhook.send(
                content=msg,
                username=f"{name} Log",
            )
        except discord.HTTPException:
            log.error(
                "Error while sending log message to webhook. Please check if the URL is correct.",
                extra={"webhook_sent": True},
            )


def _discord_filter(record):
    """A filter that blocks logs that have already been sent to a Discord webhook."""
    if "webhook_sent" in record.__dict__:
        return not record.__dict__["webhook_sent"]
    return True


def set_log(
    name: str = DEFAULT_LOG,
    log_level: int = logging.INFO,
    *,
    console: bool = True,
    file: bool | str = False,
    file_mode: str = "w",
    log_format: str | LogFormat = LogFormat.default,
    time_format: str | TimeFormat = TimeFormat.default,
    discord_log_level: int = logging.WARNING,
    webhook_url: str | None = None,
    dc_codeblocks: bool = True,
    level_spacing: int = 8,
    space_mode: Literal["auto", "always", "never"] = "auto",
    colors: dict[int, str] | str | None = None,
):
    """Creates a logger. If this logger already exists, it will return the existing logger.

    This should be called before :class:`.Bot` is initialized.

    Parameters
    ----------
    name:
        The name of the logger.
    log_level:
        The log level for default log messages ``logging.INFO``.
    console:
        Whether to log to the console. Defaults to ``True``.
    file:
        Whether to log to a file. Defaults to ``False``.
        You can also pass a path to a log file.
    file_mode:
        The file mode for the log file. Defaults to ``w``.
    log_format:
        The log format. Defaults to :attr:`.LogFormat.default`.
    time_format:
        The time format. Defaults to :attr:`.TimeFormat.default`.
    discord_log_level:
        The log level for discord log messages. Defaults to ``logging.WARNING``.

        .. note::
            For discord log messages, ``webhook_url`` is required.
    webhook_url:
        The discord webhook URL to send logs to. Defaults to ``None``.
    dc_codeblocks:
        Whether to use codeblocks for all Discord log messages. Defaults to ``True``.
    level_spacing:
        The length of the log level. If the log level is ``INFO`` and the spacing is ``8``, the log
        level will be filled with 4 additional spaces, so that all log levels are 8 characters long.
    space_mode:
        Choose when to add spacing to the log level. Defaults to ``auto``.
    colors:
        A dictionary of log levels and their corresponding colors. If only one color is given,
        all log levels will be colored with that color.

    Returns
    -------
    :class:`logging.Logger`

    Example
    -------
    .. code-block:: python

        import logging
        import ezcord

        colors = {
            logging.DEBUG: "blue",
            logging.INFO: "red",
        }

        ezcord.set_log(colors=colors)
    """
    logger = logging.getLogger(name)
    if logger.handlers:
        return logger
    logger.setLevel(log_level)

    handlers: list[logging.FileHandler | logging.StreamHandler] = []
    if isinstance(file, bool) and file:
        if not os.path.exists("logs"):
            os.mkdir("logs")
        filename = name.split(".")[-1]
        handlers.append(
            logging.FileHandler(f"logs/{filename}.log", mode=file_mode, encoding="utf-8")
        )
    elif isinstance(file, str):
        handlers.append(logging.FileHandler(file, mode=file_mode, encoding="utf-8"))

    if console:
        handlers.append(logging.StreamHandler(sys.stdout))

    space_after_level = True
    if "%(levelname)s " not in log_format and "%(levelname)s{end} " not in log_format:
        space_after_level = False
        if space_mode == "auto":
            level_spacing = 0

    if space_mode == "never":
        level_spacing = 0

    file_color_formatter = _ColorFormatter(
        True, log_format, time_format, dc_codeblocks, level_spacing, space_after_level, colors
    )
    color_formatter = _ColorFormatter(
        False, log_format, time_format, dc_codeblocks, level_spacing, space_after_level, colors
    )

    for handler in handlers:
        if isinstance(handler, logging.FileHandler):
            handler.setFormatter(file_color_formatter)
        else:
            handler.setFormatter(color_formatter)

        handler.setLevel(log_level)
        logger.addHandler(handler)

    # Discord logs
    dc_format = "**%(levelname)s:** %(message)s"
    discord_handler = _DiscordHandler(webhook_url, dc_codeblocks, dc_format)
    if dc_codeblocks:
        discord_handler.setFormatter(color_formatter)
    else:
        dc_formatter = _ColorFormatter(
            True, dc_format, time_format, dc_codeblocks, level_spacing, space_after_level, colors
        )
        discord_handler.setFormatter(dc_formatter)
    discord_handler.addFilter(_discord_filter)
    discord_handler.setLevel(discord_log_level)
    logger.addHandler(discord_handler)

    return logger



import json
import os
from functools import cache
from pathlib import Path

from ...logs import log


@cache
def load_lang(language: str) -> dict:
    """Loads the default language file and checks if the user provided a custom language file."""

    lang = {}
    parent = Path(__file__).parent.absolute()
    for element in os.scandir(parent):
        if not element.is_file() or not element.name.endswith(f"{language}.json"):
            continue

        with open(os.path.join(parent, f"{language}.json"), encoding="utf-8") as file:
            lang = json.load(file)
            break

    # check if the user has a custom language file
    for root, directories, files in os.walk(os.getcwd()):
        for filename in files:
            if filename != f"ez_{language}.json":
                continue

            log.debug(f"Custom language file loaded: **{filename}**")

            path = os.path.join(root, filename)
            with open(path, encoding="utf-8") as user_file:
                user_dic = json.load(user_file)
                for key, value in user_dic.items():
                    lang[key] = value

    if lang == {}:
        log.warn(f"Language file for language '{language}' not found. Falling back to 'en'.")

    return lang



class EzcordException(Exception):
    """Base exception class for all Ezcord exceptions."""


class MissingDiscordLibrary(EzcordException):
    """Raised when no discord library is found."""

    def __init__(self):
        super().__init__("No discord library found. Please install a supported library.")


class ConvertTimeError(EzcordException):
    """Raised when a time conversion fails."""


class Blacklisted(EzcordException):
    """Can be raised when a blacklisted user tries to use a command.

    This error can be caught in a command error handler to send a custom response.
    """



"""Utility functions for the ready event."""

from __future__ import annotations

from collections import OrderedDict
from itertools import cycle, islice

from colorama import Fore

from .. import __version__
from ..enums import ReadyEvent
from ..internal.dc import discord
from ..logs import log
from .colors import get_escape_code


class Style:
    TL = "╭"  # top left
    TR = "╮"  # top right
    BL = "╰"  # bottom left
    BR = "╯"  # bottom right
    H = "─"  # horizontal
    V = "│"  # vertical
    M = "┼"  # middle
    L = "├"  # left
    R = "┤"  # right
    T = "┬"  # top
    B = "┴"  # bottom


class Bold(Style):
    TL, TR, BL, BR, H, V, M, L, R, T, B = "╔", "╗", "╚", "╝", "═", "║", "╬", "╠", "╣", "╦", "╩"


READY_TITLE: str = f"Bot is online with EzCord {__version__}"
DEFAULT_COLORS: list[str] = [Fore.CYAN, Fore.MAGENTA, Fore.YELLOW, Fore.GREEN, Fore.BLUE, Fore.RED]


def get_default_info(bot: discord.ext.commands.Bot) -> list[tuple[str, str]]:
    try:
        lib_name = "Pycord"
        cmds = [
            cmd for cmd in bot.walk_application_commands() if type(cmd) != discord.SlashCommandGroup
        ]
    except AttributeError:
        lib_name = "Version"
        cmds = bot.commands  # prevents errors with other libs

    return [
        ("Bot", f"{bot.user}"),
        ("ID", f"{bot.user.id}"),
        (lib_name, discord.__version__),
        ("Commands", f"{len(cmds):,}"),
        ("Guilds", f"{len(bot.guilds):,}"),
        ("Latency", f"{round(bot.latency * 1000):,}ms"),
    ]


def modify_info(
    bot: discord.Bot, modifications: tuple, custom_color_list: list[str] | None = None
) -> tuple[list[tuple[str, str]], list[str]]:
    """Add or remove information from the default ready event."""

    infos = get_default_info(bot)
    additions, deletions = modifications
    colors = custom_color_list or DEFAULT_COLORS

    for key, settings in additions.items():
        value, position, color = settings["value"], settings["position"], settings["color"]

        if position is None:
            position = len(infos)

        infos.insert(position, (key, str(value)))
        colors.insert(position, get_escape_code(color))

    for key in deletions:
        if isinstance(key, int):
            try:
                infos.pop(key)
            except IndexError:
                log.warning(f"Index {key} does not exist and could not be removed.")
        else:
            infos = [info for info in infos if info[0] != key]

    return infos, colors


def print_custom_ready(
    bot: discord.Bot,
    title: str,
    modifications: tuple,
    style: ReadyEvent = ReadyEvent.default,
    default_info: bool = True,
    new_info: dict | None = None,
    colors: list[str] | None = None,
):
    if not bot.user:
        return log.error("The ready function must be called within the on_ready event.")

    colors = list(map(get_escape_code, colors or DEFAULT_COLORS))

    if default_info:
        infos, colors = modify_info(bot, modifications, colors)
    else:
        infos = get_default_info(bot)

    if new_info:
        for key, value in new_info.items():
            infos.append((str(key), str(value)))

    print_ready(bot, style, OrderedDict(infos), title, colors)


def print_ready(
    bot: discord.Bot,
    style: ReadyEvent,
    infos: OrderedDict | None = None,
    title: str = READY_TITLE,
    colors: list[str] | None = None,
    *,
    modifications: tuple | None = None,
):
    colors = colors or DEFAULT_COLORS
    infos = infos or OrderedDict(get_default_info(bot))

    if modifications:
        infos_list, colors = modify_info(bot, modifications, colors)
        infos = OrderedDict(infos_list)

    info_count = len(infos.items())
    colors = list(islice(cycle(colors), info_count))

    colon_infos = {key + ":": value for key, value in infos.items()}

    style_cls = Style()
    if "bold" in style.name:
        style_cls = Bold()

    txt = title
    if style in [ReadyEvent.box, ReadyEvent.box_bold, ReadyEvent.box_colorful]:
        txt += box(colon_infos, colors, style, style_cls)
        log.info(txt)
    elif style == ReadyEvent.logs:
        log.info(txt)
        logs(colon_infos)
    else:
        color_table = {
            key: colors[i] + value + Fore.RESET for i, (key, value) in enumerate(infos.items())
        }
        if style == ReadyEvent.table or style == ReadyEvent.table_bold:
            info_list = [list(infos.keys()), list(infos.values())]
            color_list = [list(color_table.keys()), list(color_table.values())]
        else:
            info_list = [list(item) for item in infos.items()]
            color_list = [list(item) for item in color_table.items()]

        txt += f"\n{Fore.RESET}" + tables(info_list, color_list, style_cls)
        log.info(txt)


def box(infos: dict[str, str], colors: list[str], box_style: ReadyEvent, s: Style = Style()):
    longest = max([str(i) for i in infos.values()], key=len)
    formatter = f"<{len(longest)}"
    longest_key = max([len(i) for i in infos.keys()]) + 1

    if box_style == ReadyEvent.box_colorful:
        txt = f"\n{Fore.RESET}"
    else:
        txt = "\n"

    txt += f"{s.TL}{(len(longest) + 2 + longest_key) * s.H}{s.TR}\n"
    for index, (key, info) in enumerate(infos.items()):
        key = f"{key:<{longest_key}}"
        if box_style == ReadyEvent.box_colorful:
            txt += f"{s.V} {key}{colors[index]}{info:{formatter}}{Fore.RESET} {s.V}\n"
        else:
            txt += f"{s.V} {key}{info:{formatter}} {s.V}\n"
    txt += f"{s.BL}{(len(longest) + 2 + longest_key) * s.H}{s.BR}"
    return txt


def logs(infos: dict[str, str]):
    for key, info in infos.items():
        log.info(f"{key} **{info}**")


def tables(rows: list[list[str]], color_rows: list[list[str]] | None = None, s: Style = Style()):
    """Create a table from a list of rows.

    Parameters
    ----------
    rows:
        The rows of the table. This is used to calculate the length of each column.
    color_rows:
        The rows of the table with color. This is used as the actual content of the table.
        If this is None, the content of the table will be taken from ``rows``.
    s:
        The style of the table.
    """
    if color_rows is None:
        color_rows = rows

    length = [max([len(value) for value in column]) for column in zip(*rows)]
    table = ""
    for index, (row, color_row) in enumerate(zip(rows, color_rows)):
        table += s.V

        middle_row = ""
        for max_length, content, color_content in zip(length, row, color_row):
            space_content = f" {content} "
            color_space_content = f" {color_content} "
            table += color_space_content + " " * (max_length - len(content)) + s.V
            middle_row += s.H * (max_length - len(content) + len(space_content)) + s.M

        middle_row = middle_row[:-1]
        if index == 0:
            top_row = middle_row.replace(s.M, s.T)
            table = s.TL + top_row + s.TR + "\n" + table

        if index != len(rows) - 1:
            table += "\n" + s.V + middle_row + s.V + "\n"
        else:
            bottom_row = middle_row.replace(s.M, s.B)
            table += "\n" + s.BL + bottom_row + s.BR

    return table



from __future__ import annotations

from dataclasses import dataclass


@dataclass
class Blacklist:
    db_path: str
    db_name: str
    raise_error: bool
    owner_only: bool


class EzConfig:
    """A class to store configuration values.

    These values are usually set only once, but are used throughout the runtime of the bot.
    """

    lang: str = "en"
    embed_templates: dict = {}

    # Blacklist
    admin_guilds: list[int] | None = None
    blacklist: Blacklist | None = None



"""Internal language utilities for the library."""

import inspect
from functools import cache
from pathlib import Path

from ..logs import log
from .config import EzConfig
from .language.languages import load_lang


def plural_de(amount: int, word: str, relative: bool = True) -> str:
    """Pluralize a given word (German).

    Parameters
    ----------
    amount:
        The amount to check.
    word:
        The word to pluralize.
    relative:
        Whether to use relative time. Defaults to ``True``.

    Examples
    --------
    >>> plural_de(1, "Tag")
    'Tag'
    >>> plural_de(2, "Tag")  # Relative: "Seit 5 Tagen"
    'Tagen'
    >>> plural_de(2, "Tag", relative=False)  # Not relative: "5 Tage"
    'Tage'
    """
    if amount != 1:
        if not word.endswith("e"):
            word += "e"

        if relative:
            return f"{word}n"

    return word


def plural_en(amount: int, word: str) -> str:
    """Pluralize a given word (English).

    Parameters
    ----------
    amount:
        The amount to check.
    word:
        The word to pluralize.

    Returns
    -------
    :class:`str`
        The pluralized word.
    """
    if amount != 1:
        return f"{word}s"
    return word


def plural_es(amount: int, word: str) -> str:
    """Pluralize a given word in Spanish.

    Parameters
    ----------
    amount:
        The amount to check.
    word:
        The word to pluralize.

    Returns
    -------
    :class:`str`
        The pluralized word.
    """
    if amount != 1:
        if word.endswith("ión"):
            return f"{word[:-3]}iones"
        elif word.endswith("z"):
            return f"{word[:-1]}ces"
        else:
            return f"{word}s"
    return word


def plural_fr(amount: int, word: str) -> str:
    """Pluralize a given word in French.

    Parameters
    ----------
    amount:
        The amount to check.
    word:
        The word to pluralize.

    Returns
    -------
    :class:`str`
        The pluralized word.
    """
    if amount != 1:
        if word.endswith("al"):
            return f"{word}aux"
        elif word.endswith("ail"):
            return f"{word}s"
        else:
            return f"{word}s"
    return word


def tp(key: str, amount: int, *args: str, relative: bool = True) -> str:
    """Load a string in the selected language and pluralize it.

    Parameters
    ----------
    key:
        The text to load.
    amount:
        The amount to check.
    *args:
        The arguments to format the string with.
    relative:
        Whether to use relative time. Defaults to ``True``.
    """
    word = t(key, *args)
    lang = get_lang()

    if lang == "de":
        return plural_de(amount, word, relative)
    elif lang == "es":
        return plural_es(amount, word)
    elif lang == "fr":
        return plural_fr(amount, word)
    else:
        return plural_en(amount, word)


def t(key: str, *args: str):
    """Load a string in the selected language.

    Parameters
    ----------
    key:
        The text to load.
    *args:
        The arguments to format the string with.
    """
    n = 1
    origin_file = Path(inspect.stack()[n].filename).stem

    while origin_file == Path(__file__).stem:
        n += 1
        origin_file = Path(inspect.stack()[n].filename).stem

    lang = get_lang()

    try:
        string = load_lang(lang)[origin_file][key]
        if not string:
            return None
        return string.format(*args)
    except KeyError:
        # fallback to english if the key is not in the custom language file
        # provided by the user
        log.warn(f"Key '{key}' not found in language file '{lang}'. Falling back to 'en'.")
        return load_lang("en")[origin_file][key].format(*args)


@cache
def get_lang():
    """Get the language from the config class."""
    return EzConfig.lang


def set_lang(lang: str):
    """Set the language for the bot.

    Parameters
    ----------
    lang:
        The language to set.
    """
    EzConfig.lang = lang



"""Some utility functions for Pycord and Python."""

from __future__ import annotations

import asyncio
import io
import itertools
import json
import random
from typing import Any

from .internal import get_lang
from .internal.dc import AutocompleteFunc, V, Values, discord


def create_json_file(
    dictionary: dict,
    filename: str = "data.json",
    *,
    indent: int = 2,
    description: str | None = None,
    spoiler: bool = False,
    **kwargs,
) -> discord.File:
    """Create a :class:`discord.File` object from a dictionary.

    Parameters
    ----------
    dictionary:
        The dictionary to convert to a JSON file.
    filename:
        The filename to use for the JSON file.
    indent:
        The indent to use for the JSON file.
    description:
        The description to use for the discord file.
    spoiler:
        Whether the Discord file should be a spoiler.
    **kwargs:
        Additional keyword arguments for :py:func:`json.dumps`.

    Returns
    -------
    :class:`discord.File`
    """
    content = json.dumps(dictionary, indent=indent, **kwargs).encode()
    return discord.File(
        io.BytesIO(content),
        filename=filename,
        description=description,
        spoiler=spoiler,
    )


def create_text_file(text: str, filename: str = "data.txt", **kwargs) -> discord.File:
    """Create a :class:`discord.File` object from a string.

    Parameters
    ----------
    text:
        The string to convert to a text file.
    filename:
        The filename to use for the text file.
    **kwargs:
        Keyword arguments for :class:`discord.File`.

    Returns
    -------
    :class:`discord.File`
    """
    return discord.File(io.BytesIO(text.encode()), filename=filename, **kwargs)


def create_html_file(html: str, filename: str = "data.html", **kwargs) -> discord.File:
    """Create a :class:`discord.File` object from an HTML string.

    Parameters
    ----------
    html:
        The HTML string to convert to an HTML file.
    filename:
        The filename to use for the HTML file.
    **kwargs:
        Keyword arguments for :class:`discord.File`.

    Returns
    -------
    :class:`discord.File`
    """
    return create_text_file(html, filename, **kwargs)


def random_avatar() -> str:
    """Returns the URL of a random default avatar."""

    return f"https://cdn.discordapp.com/embed/avatars/{random.randint(0, 5)}.png"


def codeblock(content: int | str, *, lang: str = "yaml", unit: str = ""):
    """Returns a codeblock with the given content.

    Parameters
    ----------
    content:
        The content of the codeblock. If the content is an integer, it will be
        formatted with commas (or dots if the language is German).
    lang:
        The language of the codeblock. Defaults to ``yaml``.
    unit:
        The text to display after the given content. This is only used if the content is an integer.
    """

    if isinstance(content, int):
        number = f"{content:,}"
        if get_lang() == "de":
            number = number.replace(",", ".")
        block = f"```{lang}\n{number}"
        if unit:
            block += f" {unit}"
        block += "```"
        return block

    return f"```{lang}\n{content}```"


def ez_autocomplete(values: Values) -> AutocompleteFunc:
    """A similar function as :func:`~discord.utils.basic_autocomplete`, but instead of returning
    options starting with the user's value, it returns options containing the user's value.

    Parameters
    ----------
    values: Iterable[str]
        Accepts an iterable of :class:`str`, a callable (sync or async) that takes a single argument
        of :class:`~discord.AutocompleteContext`, or a coroutine. Must resolve to an iterable of :class:`str`.
    """

    async def autocomplete_callback(ctx: discord.AutocompleteContext) -> V:
        _values = values

        if callable(_values):
            _values = _values(ctx)
        if asyncio.iscoroutine(_values):
            _values = await _values

        def check(item: Any) -> bool:
            item = getattr(item, "name", item)
            return str(ctx.value or "").lower() in str(item).lower()

        gen = (val for val in _values if check(val))
        return iter(itertools.islice(gen, 25))

    return autocomplete_callback
